# User defined errors
class Error(Exception):
    """Base class for exceptions in this module."""

    pass


class InputError(Error):
    """Exception raised for errors in the input."""

    name = "InputError"

    def __init__(self, error_ctx):
        self.error_ctx = error_ctx


class ParseError(Error):
    """Exception raised for errors in the configuration file."""

    name = "ParseError"

    def __init__(self, error_ctx):
        self.error_ctx = error_ctx


class ConfigError(Error):
    """Exception raised for empty configuration values."""

    name = "ConfigError"

    def __init__(self, error_ctx):
        self.error_ctx = error_ctx


class ApiError(Error):
    """Exception raised when trying to run the application
    without API keys."""

    name = "ApiError"

    def __init__(self, error_ctx):
        self.error_ctx = error_ctx



import os
from pathlib import Path
import yaml
import click
import logging
from imgdb.utils import Tcolors, logger
from imgdb.exceptions import ParseError, ConfigError, ApiError


class Config:
    """Default configuration."""

    DEV_MODE = os.getenv("IMGDB_DEV")
    HOME = Path("./config_test").resolve() if DEV_MODE else Path.home()
    XDG_CONFIG_HOME = Path(
        os.getenv("XDG_CONFIG_HOME")
        if os.getenv("XDG_CONFIG_HOME")
        else HOME / ".config"
    )
    XDG_DATA_HOME = Path(
        os.getenv("XDG_DATA_HOME")
        if os.getenv("XDG_DATA_HOME")
        else HOME / ".local/share"
    )
    XDG_CACHE_HOME = Path(
        os.getenv("XDG_CACHE_HOME") if os.getenv("XDG_CACHE_HOME") else HOME / ".cache"
    )
    IMGDB_CONFIG_HOME = Path(XDG_CONFIG_HOME / "imgdb")
    IMGDB_DATA_HOME = Path(XDG_DATA_HOME / "imgdb")
    IMGDB_CACHE_HOME = Path(XDG_CACHE_HOME / "imgdb")
    CONFIG = Path(IMGDB_CONFIG_HOME / "imgdb.yaml")
    IMDB_DATASETS_DIR = Path(IMGDB_DATA_HOME / "imdb_datasets")

    DEFAULT_CONFIG = {
        "general": {
            "update frequency": "weekly",
            "download": False,
            "image editing": False,
            "log file path": str(Path(IMGDB_CACHE_HOME / "imgdb.log")),
        },
        "interface": {
            "api": {
                "google search api key": "",
                "imdb custom search id": "",
            }
        },
    }


def check_config_file(debug):
    """This function checks if a config file exists
    in the CONFIG directory."""

    # Preparing the necessary configuration & data directories
    imdb_datasets_dir = (Config.IMDB_DATASETS_DIR, "Imdb datasets")
    imgdb_cache_dir = (Config.IMGDB_CACHE_HOME, "cache")
    dir_list = [imdb_datasets_dir, imgdb_cache_dir]
    for d in dir_list:
        if not d[0].is_dir():
            try:
                d[0].mkdir(parents=True)
            except Exception as e:
                click.echo(
                    Tcolors.FAIL + "Something went wrong while trying"
                    " to create the '%s' directory!" % d[1] + Tcolors.ENDC
                )
                return 0

    logger(
        debug,
        Config.DEFAULT_CONFIG["general"]["log file path"],
    )

    config_file = Config.CONFIG
    if config_file.is_file():
        return read_config_yaml(config_file)
    else:
        return create_default_yaml(config_file)


def create_default_yaml(config_file):
    """This function creates and saves the default configuration file."""

    config_file_path = config_file
    imgdb_config_dir = Config.IMGDB_CONFIG_HOME

    if not imgdb_config_dir.is_dir():
        try:
            imgdb_config_dir.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            logging.critical(
                "Something went wrong while trying to create"
                " the configuration directory!"
            )
            logging.debug("Error: %s" % e)
            click.echo(
                Tcolors.FAIL + "Something went wrong while trying to create"
                " the configuration directory!" + Tcolors.ENDC
            )
            return 0

    try:
        with open(config_file_path, "w", encoding="utf-8") as config:
            yaml.safe_dump(
                Config.DEFAULT_CONFIG,
                config,
                encoding="utf-8",
                allow_unicode=True,
                default_flow_style=False,
            )
        click.echo(
            Tcolors.OK_GREEN
            + "➜ The configuration file: %s \nhas been created successfully!"
            % config_file_path
            + Tcolors.ENDC
        )
    except Exception as e:
        logging.critical(
            "Something went wrong while trying to save"
            " the program's configuration file!"
        )
        logging.debug("Error: %s" % e)
        click.echo(
            Tcolors.FAIL + "Something went wrong while trying to save"
            " the program's configuration file!" + Tcolors.ENDC
        )
        return 0
    return parse_config_yaml(Config.DEFAULT_CONFIG, first_run=True)


def read_config_yaml(config_file):
    """This function extracts configuration options
    from a YAML configuration file."""

    config_file_path = config_file
    try:
        with open(config_file_path, "r") as config:
            current_config = yaml.safe_load(config)
            logging.debug("Current configuration dictionary: %s" % current_config)
            return parse_config_yaml(current_config)
    except yaml.YAMLError as exc:
        if hasattr(exc, "problem_mark"):
            mark = exc.problem_mark
            logging.critical(
                "The configuration file is malformatted."
                " Error(s) at line %s, column %s!" % (mark.line + 1, mark.column + 1)
            )
            click.echo(
                Tcolors.FAIL + "The configuration file is malformatted."
                " Error(s) at line %s, column %s!" % (mark.line + 1, mark.column + 1)
                + Tcolors.ENDC
            )
        else:
            logging.critical("Failed to parse the configuratiom file!")
            click.echo(
                Tcolors.FAIL + "Failed to parse the configuratiom file!" + Tcolors.ENDC
            )
        return 0
    except Exception as e:
        if hasattr(e, "error_ctx"):
            logging.critical("Error: %s. Context: %s" % (e.name, e.error_ctx))
            click.echo(Tcolors.FAIL + e.error_ctx + Tcolors.ENDC)
        else:
            logging.critical(
                "Something went wrong while trying to read"
                " the program's configuration file!"
            )
            logging.debug("Error: %s" % e)
            click.echo(
                Tcolors.FAIL + "Something went wrong while trying to read"
                " the program's configuration file!" + Tcolors.ENDC
            )
        return 0


def parse_config_yaml(current_config, first_run=False):
    """This function parses the configuration values contained in the
    configuration dictionary obtained by reading the YAML file."""

    config_options = current_config
    used_options = {}

    if config_options is None:
        return validate_config(
            cfg_error=True, cfg_error_ctx="The configuration file cannot be empty!"
        )

    # Aggressively parsing the current configuration options allowing
    # neither empty nor unknown options
    for key, value in config_options.items():
        if key == "general" and isinstance(value, dict):
            for gen_key, gen_value in config_options[key].items():
                if gen_key == "log file path" and not isinstance(gen_value, dict):
                    used_options[gen_key] = gen_value
                elif gen_key == "download" and not isinstance(gen_value, dict):
                    used_options[gen_key] = gen_value
                elif gen_key == "image editing" and not isinstance(gen_value, dict):
                    used_options[gen_key] = gen_value
                elif gen_key == "font" and not isinstance(gen_value, dict):
                    used_options[gen_key] = gen_value
                elif gen_key == "update frequency" and not isinstance(gen_value, dict):
                    used_options[gen_key] = gen_value
                else:
                    raise ParseError(
                        "Unknown or malformatted option <%s> in %s section!"
                        % (gen_key, key)
                        if not isinstance(gen_value, dict)
                        else "Invalid value <%s> for option <%s> in '%s' section!"
                        % (gen_value, gen_key, key)
                    )

        elif key == "interface" and isinstance(config_options[key], dict):
            for inter_key, inter_value in value.items():
                if inter_key == "api" and isinstance(inter_value, dict):
                    for api_key, api_value in value[inter_key].items():
                        if api_key == "google search api key" and not isinstance(
                            api_value, dict
                        ):
                            used_options[api_key] = api_value
                        elif api_key == "imdb custom search id" and not isinstance(
                            api_value, dict
                        ):
                            used_options[api_key] = api_value
                        else:
                            raise ParseError(
                                "Unknown or malformatted option <%s> in %s section!"
                                % (api_key, key)
                                if not isinstance(api_value, dict)
                                else "Invalid value <%s> for option <%s> in '%s' section!"
                                % (api_value, api_key, key)
                            )
                else:
                    raise ParseError(
                        "Unknown or malformatted option <%s> in '%s' section!"
                        % (inter_key, key)
                    )
        else:
            raise ParseError(
                "Unknown or malformatted section <%s> in the configuration file!" % key
            )
    return validate_config(used_options, check_api=first_run)


def validate_config(
    config_options={}, check_api=False, cfg_error=None, cfg_error_ctx=None
):
    """This function checks if configuration values are not empty or null."""

    if cfg_error and cfg_error_ctx:
        raise ConfigError(cfg_error_ctx)
        return 0

    try:
        if check_api:
            raise ApiError(
                "You are running the application for the first time."
                " Please, provide a working Google Search API key in the configuration file!"
            )
    except ApiError as e:
        logging.critical("Error: %s. Context: %s" % (e.name, e.error_ctx))
        click.echo(Tcolors.FAIL + e.error_ctx + Tcolors.ENDC)
        return 0

    api_keys = ["google search api key", "imdb custom search id"]
    for key in api_keys:
        try:
            if key not in config_options:
                raise ApiError(
                    "You must provide a working Google Custom Search"
                    " API key in the configuration file!"
                )
                break
        except ApiError as e:
            logging.critical("Error: %s. Context: %s" % (e.name, e.error_ctx))
            click.echo(Tcolors.FAIL + e.error_ctx + Tcolors.ENDC)
            return 0

    for key, value in config_options.items():
        try:
            if value is None or value == "":
                raise ConfigError(
                    "The <%s> option in the configuration file cannot be empty!" % key
                )
        except ConfigError as e:
            logging.critical("Error: %s. Context: %s" % (e.name, e.error_ctx))
            click.echo(Tcolors.FAIL + e.error_ctx + Tcolors.ENDC)
            return 0
    return config_options


if __name__ == "__main__":
    # print(Path(IMGDB_DATA_HOME / "imgdb.log"))
    # a = check_config_file()
    # print("Returned value is: %s" % a)
    # create_default_yaml(Config.CONFIG)
    print("DEV", Config.DEV_MODE)
    print("XDG HOME", Config.XDG_CONFIG_HOME)
    print("XDG DATA HOME", Config.XDG_DATA_HOME)
    print("IMGDB CONFIG HOME", Config.IMGDB_CONFIG_HOME)
    print("IMGDB DATA HOME", Config.IMGDB_DATA_HOME)
    print("CONFIG", Config.CONFIG)
    print("DATASETS DIR", Config.IMDB_DATASETS_DIR)



from urllib.request import urlopen
from urllib.error import URLError, HTTPError
from tqdm import tqdm
from imgdb.utils import Tcolors
import logging
import click
import re
import sys


def imdb_download_poster(
    url, name=None, filepath=None, download_or_update="Downloading"
):
    """This function downloads a file from a given URL."""

    file_url = url
    url_test = "http://download.thinkbroadband.com/10MB.zip"
    if not filepath:
        naked_file_name = name
        ext = file_url.split(".")[-1]

        special_letters = "àáâäæãåāăąçćčđďèéêëēėęěğǵḧîïíīįìıİłḿñńǹňôöòóœøōõőṕŕřßśšşșťțûüùúūǘůűųẃẍÿýžźż"
        normal_letters = "aaaaaaaaaacccddeeeeeeeegghiiiiiiiilmnnnnoooooooooprrsssssttuuuuuuuuuwxyyzzz"
        regex_special_letters = r"|".join(list(special_letters))
        regex_special_characters = r"[\\~#%&*{}\/:,;<>?!|\"-.\s]{1,}"
        regex_special_letters_test = re.search(regex_special_letters, naked_file_name)

        if regex_special_letters_test:
            name_no_special_letters = re.sub(
                regex_special_letters,
                lambda reg: list(normal_letters)[
                    list(special_letters).index(reg.group())
                ],
                naked_file_name,
            )
        else:
            name_no_special_letters = naked_file_name

        name_no_special_characters = re.sub(
            regex_special_characters, "_", name_no_special_letters
        )

        file_name = name_no_special_characters + "." + ext
        file_path = file_name
    else:
        file_path = filepath
        file_name = file_url.split("/")[-1]

    try:
        r = urlopen(url)
    except HTTPError as e:
        logging.critical("Download server couldn't fulfill the request.")
        logging.debug("Error code: %s" % e.code)
        click.echo(
            Tcolors.FAIL
            + "Download server couldn't fulfill the request."
            + Tcolors.ENDC
        )
        sys.exit()
    except URLError as e:
        logging.critical("We failed to reach the download server.")
        click.echo(
            Tcolors.FAIL + "We failed to reach the download server." + Tcolors.ENDC
        )
        if hasattr(e, "reason"):
            logging.debug("Reason: %s" % e.reason)
        sys.exit()

    else:
        try:
            file_size = int(r.getheader("Content-Length"))
        except (TypeError, ValueError):
            logging.critical("Filesize has to be an integer!")
            click.echo(Tcolors.FAIL + "Filesize has to be an integer!" + Tcolors.ENDC)
            sys.exit()

        logging.info(
            "%s: %s - Size: %s KiB" % (download_or_update, file_name, file_size // 1024)
        )
        click.echo(
            Tcolors.OK_GREEN
            + "➜ %s: %s - Size: %s KiB\n"
            % (download_or_update, file_name, file_size // 1024)
            + Tcolors.ENDC
        )

        block_size = 1024
        progress_bar = tqdm(total=file_size, unit="iB", unit_scale=True)
        with open(file_path, "wb") as f:
            while True:
                buffer = r.read(block_size)
                if not buffer:
                    break

                f.write(buffer)
                progress_bar.update(len(buffer))
        progress_bar.close()

        if file_size != 0 and progress_bar.n != file_size:
            logging.critical("It looks like something unexpected happened. Aborting!")
            click.echo(
                Tcolors.FAIL
                + "It looks like something unexpected happened. Aborting!"
                + Tcolors.ENDC
            )
            sys.exit()

        if not filepath:
            saved_file_data = {
                "filename": name_no_special_characters,
                "filepath": file_path,
            }
            return saved_file_data


if __name__ == "__main__":
    f = imdb_download_poster(
        "https://m.media-amazon.com/images/M/MV5BYjc1ZTFiNGItMzQyYy00OTFlLThjOGYtNzA2NWY1M2E4MTAzXkEyXkFqcGdeQXVyMTkxNjUyNQ@@._V1_FMjpg_UX1000_.jpg",
        name="Dunnu éla#dopeáj ffrg",
    )
    print(f)



import datetime
import json
from urllib.request import urlopen, Request
import urllib.parse
from bs4 import BeautifulSoup
from random import uniform, choice
from time import sleep
import re
from urllib.error import URLError, HTTPError
import os
from imgdb.imdb_data_handler import (
    imdb_get_data_from_datasets,
    merge_tsv_files,
    datasets_updater,
)
from imgdb.imdb_poster_fetcher import imdb_download_poster
from imgdb.image_processor import generate_media_image
import click
from imgdb.utils import *
from imgdb.exceptions import InputError
from imgdb.config import check_config_file, Config
from pathlib import Path
from difflib import SequenceMatcher


@click.command()
@click.option("--mov", help="The title of the movie.")
@click.option("--tv", help="The title of the series.")
@click.option("--tvmini", help="The title of the mini series.")
@click.option("--debug", help="The logging level of the application.")
@click.option("--logfile", help="The path of the log file.")
@click.option("--freq", help="The update frequency of the datasets.")
@click.option("--font", help="The font used to generate the rating image.")
@click.option(
    "-d", is_flag=True, default=False, help="Download the movie's poster image."
)
@click.option(
    "-e",
    is_flag=True,
    default=False,
    help="Save the edited image containing the movie's ratings.",
)
def imdb_cli_init(mov, tv, tvmini, debug, logfile, freq, font, d, e):
    """Imdb CLI search."""

    # Checking and creating or getting current config info
    debug = "debug" if Config.DEV_MODE else "warning"
    current_config = check_config_file(debug)
    if current_config == 0:
        return

    # Making sure that command-line options override those
    # present in the configuration file
    runtime_options = {
        "download": d if d else current_config.get("download"),
        "image_edit": e if e else current_config.get("image editing"),
        "font": font if font else current_config.get("font"),
        "log_file_path": logfile if logfile else current_config.get("log_file_path"),
        "freq": freq if freq else current_config.get("update frequency"),
        "gsearch_api_key": current_config.get("google search api key"),
        "imdb_gsearch_id": current_config.get("imdb custom search id"),
    }
    logging.debug("Runtime options: %s" % runtime_options)

    # Setting up a logger
    if runtime_options["log_file_path"]:
        if runtime_options["log_file_path"].split(".")[-1] == "log":
            log_file_path = Path(runtime_options["log_file_path"]).resolve()
            logfile_directory = log_file_path.parents[0]
            is_logfile = log_file_path.is_file()
            is_logfile_path = logfile_directory.is_dir()
            test_log_paths = (
                str(log_file_path) == Config.DEFAULT_CONFIG["general"]["log file path"]
            )

            if not test_log_paths:
                if is_logfile_path:
                    try:
                        os.link(
                            Config.DEFAULT_CONFIG["general"]["log file path"],
                            log_file_path,
                        )
                    except PermissionError:
                        click.echo(
                            Tcolors.FAIL + "Permission denied. You don't have"
                            " the required permissions on '%s'!" % logfile_directory
                            + Tcolors.ENDC
                        )
                        logging.critical(
                            "Permission denied. You don't have"
                            " the required permissions on '%s'!" % logfile_directory
                        )
                        return
                else:
                    click.echo(
                        Tcolors.WARNING
                        + "The log file's directory does not exist!"
                        + Tcolors.ENDC
                    )
                    logging.warning("The log file's directory does not exist!")
                    return
        else:
            click.echo(
                Tcolors.WARNING + "Either the log file's path is invalid or that"
                " it doesn't end with the .log extension!" + Tcolors.ENDC
            )
            logging.warning(
                "Either the log file's path is invalid or that"
                " it doesn't end with the .log extension!"
            )
            return

    # Updating the datasets
    dataset_update_status = datasets_updater(runtime_options["freq"])
    if dataset_update_status == 0:
        return

    # Creating a dictionary containing a list of all mutually exclusive options
    options = {"mov": mov, "tv": tv, "tvmini": tvmini}

    used_options = []

    # Retrieving all used options and storing them in a list
    for key, value in options.items():
        if options[key] != None:
            used_options.append(options[key])
    try:
        if len(used_options) == 0:
            raise InputError("Command launched without options!")
    except InputError as e:
        logging.warning("Error: %s. Context: %s" % (e.name, e.error_ctx))
        click.echo(
            Tcolors.WARNING + "Please specify the media type argument!" + Tcolors.ENDC
        )
    else:
        # Here, this code only executes when the user has specified only
        # one option from the mutually exclusive options
        if len(used_options) == 1:
            media_name = used_options[0]
            media_type = list(options.keys())[list(options.values()).index(media_name)]

            logging.debug("Media title: %s" % media_name)
            logging.debug("Media type: %s" % media_type)
            click.echo(Tcolors.OK_GREEN + "Fetching data..." + Tcolors.ENDC)

            imdb_data = imdb_get_data(
                media_name,
                media_type,
                api_keys=[
                    runtime_options["gsearch_api_key"],
                    runtime_options["imdb_gsearch_id"],
                ],
            )

            if imdb_data:
                rt_data = rt_get_data(
                    imdb_data["imdb_title"],
                    imdb_data["imdb_original_title"],
                    media_type,
                    imdb_data["imdb_year"],
                )
                rt_media_rating = rt_data["rt_rating"] if rt_data else "N/A"

                click.echo(
                    "Title: %s" % imdb_data["imdb_title"]
                    + "\nGenres: %s" % ", ".join(map(str, imdb_data["imdb_genres"]))
                    + "\nYear: %s" % imdb_data["imdb_year"]
                    + "\nRuntime: %s min" % imdb_data["imdb_runtime"]
                    + "\nIMDb Rating: %s" % imdb_data["imdb_rating"]
                    + "\nRottenTomatoes Rating: %s" % rt_media_rating
                    + "\nDescription: %s"
                    % replace_every_nth(50, " ", "\n", imdb_data["imdb_description"])
                )

                if (
                    runtime_options["download"] == True
                    or runtime_options["image_edit"] == True
                ):
                    downloaded_image_data = imdb_download_poster(
                        imdb_data["imdb_poster_url"], imdb_data["imdb_title"]
                    )
                    if runtime_options["image_edit"] == True:
                        generate_media_image(
                            imdb_data["imdb_title"],
                            imdb_data["imdb_rating"],
                            rt_media_rating,
                            downloaded_image_data["filename"],
                            downloaded_image_data["filepath"],
                            font=runtime_options["font"],
                        )
                    elif runtime_options["image_edit"]:
                        click.echo(
                            Tcolors.FAIL
                            + "The image editing option is invalid!"
                            + Tcolors.ENDC
                        )
                        logging.critical("The image editing option is invalid!")

                elif runtime_options["download"]:
                    click.echo(
                        Tcolors.FAIL + "The download option is invalid!" + Tcolors.ENDC
                    )
                    logging.critical("The download option is invalid!")

        else:
            # The user has specified at least two mutually exclusive options
            mut_exclusive_options = [
                ex for ex in options.keys() if options[ex] is not None
            ]
            click.echo(
                Tcolors.WARNING
                + " and ".join(mut_exclusive_options)
                + " are conflicting options. Please use only one option at a time!"
                + Tcolors.ENDC
            )
            logging.warning(
                " and ".join(mut_exclusive_options)
                + " are conflicting options. Please use only one option at a time!"
            )


def rt_construct_json():
    """This function creates a JSON file that contains a list
    of movies with their respective metadata."""

    data = rt_parse_json()
    final_data = []

    for item in data:
        logging.info("Here we go! The movie's title is: %s" % item["title"])
        # Random wait to avoid rate limits
        sleep(uniform(1, 2.5))
        imdb_results = imdb_get_data(item["title"])

        movie_data = {
            "tomatoTitle": item["title"],
            "tomatoScore": item["tomatoScore"],
            "imdbRating": imdb_results["imdb_rating"],
            "dvdReleaseDate": item.get("dvdReleaseDate")
            if item.get("dvdReleaseDate")
            else "N/A",
            "year": rt_get_movie_year(item["url"]),
            "runtime": item.get("runtime") if item.get("runtime") else "N/A",
            "mpaaRating": item.get("mpaaRating") if item.get("mpaaRating") else "N/A",
            "thumbnailUrl": imdb_results["imdb_thumbnail_url"],
            "posterUrl": imdb_results["imdb_poster_url"],
        }

        final_data.append(movie_data.copy())

    # Writing the resulting dictionary to a JSON file
    write_json_to_file(final_data)


def rt_parse_json():
    """This is a function that fetches and parses JSON data from
    a url (RottenTomatoes semi-public API)."""

    rt_base_url = "https://www.rottentomatoes.com/api/private/v2.0"
    api_urls = [
        f"{rt_base_url}/browse?minTomato=70&maxTomato=100"
        "&maxPopcorn=100&services=amazon%3Bhbo_go%3B"
        "itunes%3Bnetflix_iw%3Bvudu%3Bamazon_prime%3Bfandango_now"
        "&certified=true&sortBy=release&type=cf-dvd-streaming-all",
        f"{rt_base_url}/browse?minTomato=70&maxTomato=100"
        "&maxPopcorn=100&services=amazon%3Bhbo_go%3B"
        "itunes%3Bnetflix_iw%3Bvudu%3Bamazon_prime%3Bfandango_now"
        "&certified=true&sortBy=release&type=cf-dvd-streaming-all&page=2",
    ]
    full_results = []

    for url in api_urls:
        try:
            res = urlopen(url)
        except HTTPError as e:
            logging.critical("RT server couldn't fulfill the request.")
            logging.debug("Error code: %s" % e.code)
            click.echo(
                Tcolors.FAIL + "RT server couldn't fulfill the request." + Tcolors.ENDC
            )
        except URLError as e:
            logging.critical("We failed to reach RT server.")
            click.echo(Tcolors.FAIL + "We failed to reach RT server." + Tcolors.ENDC)
            if hasattr(e, "reason"):
                logging.debug("Reason: %s" % e.reason)

        else:
            data = json.loads(res.read().decode())
            results = data["results"]
            full_results.extend(results)
    return full_results


def rt_search_media(title, mtype):
    """This is a function that searches RT site for a specific title (movie or series)."""

    rt_search_term = title
    media_type = mtype
    query = urllib.parse.quote_plus(rt_search_term)
    url = f"https://www.rottentomatoes.com/search?search={query}"

    # Implementing a random wait timer to avoid some anti-scraping detection
    # methods (not sure if it works but gonna keep it anyway).
    sleep(uniform(0.05, 0.1))

    # Creating a list of user agents to use with urllib
    user_agents_list = [
        "Mozilla/5.0 (Windows NT 5.1; rv:11.0) Gecko Firefox/11.0"
        " (via ggpht.com GoogleImageProxy)",
        "CheckMarkNetwork/1.0 (+http://www.checkmarknetwork.com/spider.html)",
        "Mozilla/5.0 (iPhone; CPU iPhone OS 13_3_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML,"
        " like Gecko) Mobile/15E148 [FBAN/FBIOS;FBDV/iPhone11,8;FBMD/iPhone;FBSN/iOS;FBSV/13.3.1;"
        "FBSS/2;FBID/phone;FBLC/en_US;FBOP/5;FBCR/]",
        "Mozilla/5.0 CK={} (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko)"
        " Chrome/52.0.2743.116 Safari/537.36 Edge/15.15063",
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko)"
        " Chrome/84.0.4147.105 Safari/537.36",
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/605.1.15"
        " (KHTML, like Gecko) Version/11.1.2 Safari/605.1.15",
        "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko)"
        " Chrome/80.0.3987.87 Safari/537.36",
        "Mozilla/5.0 (Linux; Android 7.1.2; AFTMM Build/NS6265; wv) AppleWebKit/537.36"
        " (KHTML, like Gecko) Version/4.0 Chrome/70.0.3538.110 Mobile Safari/537.36",
    ]

    # Picking up a random user agents each time we make a connection to the server
    headers = {
        "User-Agent": choice(user_agents_list),
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
        "Accept-Language": "en-US,en;q=0.5",
        "DNT": "1",
        "Connection": "keep-alive",
        "Upgrade-Insecure-Requests": "1",
    }

    req = Request(url, headers=headers)

    try:
        response = urlopen(req)
    except HTTPError as e:
        logging.critical("RT server couldn't fulfill the request.")
        logging.debug("Error code: %s" % e.code)
        click.echo(
            Tcolors.FAIL + "RT server couldn't fulfill the request." + Tcolors.ENDC
        )
    except URLError as e:
        logging.critical("We failed to reach RT server.")
        click.echo(Tcolors.FAIL + "We failed to reach RT server." + Tcolors.ENDC)
        if hasattr(e, "reason"):
            logging.debug("Reason: %s" % e.reason)

    else:
        soup = BeautifulSoup(response.read(), "html.parser")
        try:
            if media_type == "movies":
                rt_media_title = soup.find(
                    "search-page-result", {"type": "movie"}
                ).find("a", {"data-qa": "info-name"})
            elif media_type == "tvSeries":
                rt_media_title = soup.find("search-page-result", {"type": "tv"}).find(
                    "a", {"data-qa": "info-name"}
                )
        except Exception as e:
            logging.critical("We couldn't retrieve the movie's title from RT.")
            logging.debug("Error: %s" % e)
            click.echo(
                Tcolors.FAIL
                + "We couldn't retrieve the movie's title from RT."
                + Tcolors.ENDC
            )
            return
        logging.info("RT media title is %s: " % rt_media_title.text.strip())
        return rt_media_title.text.strip()


def imdb_get_data(title, mtype, api_keys=[]):
    """This function fetches the poster url and other data related
    to Imdb movies from Google Custom Search JSON API."""

    movie_title = title
    media_type = mtype
    gcsearch_api_key = api_keys[0]
    imdb_custom_search_id = api_keys[1]
    query = urllib.parse.quote_plus(movie_title)
    api_base_url = "https://www.googleapis.com/customsearch/v1/siterestrict?key="
    url = (
        f"{api_base_url}{gcsearch_api_key}&cx={imdb_custom_search_id}&num=10&q={query}"
    )
    regex_url = r"(?<=UY1200)(.*?)(?=.(jpg|png|jpeg)\b)"
    regex_year_parentheses = r"\(([^)]+)\)"
    regex_year = r"\d+"
    regex_title = r"(^.*?)(?=\s\()"

    try:
        res = urlopen(url)
    except HTTPError as e:
        logging.critical("Google Search server couldn't fulfill the request.")
        logging.debug("Error code: %s" % e.code)
        click.echo(
            Tcolors.FAIL
            + "Google Search server couldn't fulfill the request."
            + Tcolors.ENDC
        )
    except URLError as e:
        logging.critical("We failed to reach Google Search server.")
        click.echo(
            Tcolors.FAIL + "We failed to reach Google Search server." + Tcolors.ENDC
        )
        if hasattr(e, "reason"):
            logging.debug("Reason: %s" % e.reason)

    else:
        data = json.loads(res.read().decode())
        item_list = data.get("items")
        results = {}

        if item_list:
            for item in item_list:
                try:
                    imdb_title = item["title"]
                    # imdb_rating = item["pagemap"]["aggregaterating"][0]["ratingvalue"]
                    imdb_description = item["pagemap"]["metatags"][0]["og:description"]
                    imdb_movie_thumbnail_url = item["pagemap"]["cse_thumbnail"][0][
                        "src"
                    ]
                    imdb_movie_cropped_poster_url = item["pagemap"]["metatags"][0][
                        "og:image"
                    ]
                except KeyError as err:
                    logging.critical("KeyError: {0}".format(err))
                    click.echo(Tcolors.FAIL + "KeyError: %s" % err + Tcolors.ENDC)
                    continue
                except NameError as err:
                    logging.critical("NameError: {0}".format(err))
                    click.echo(Tcolors.FAIL + "NameError: %s" % err + Tcolors.ENDC)
                except IndexError as err:
                    logging.critical("IndexError: {0}".format(err))
                    click.echo(Tcolors.FAIL + "IndexError: %s" % err + Tcolors.ENDC)
                except Exception as e:
                    logging.critical("Unexpected error: {0}".format(e))
                    click.echo(Tcolors.FAIL + "Unexpected error: %s" % e + Tcolors.ENDC)
                    raise
                else:

                    # Google custom search engine changed their JSON response.
                    # Now, there is no need to process the imdb_movie_cropped_poster_url.
                    # The imdb_movie_cropped_poster_url which we get directly from
                    # the JSON results already contains the desired image size.
                    # Thus, there is no need to use regex substitutions on
                    # imdb_movie_cropped_poster_url.
                    # But I will leave the part that deals with substitutions intact
                    # as Google can change their API anytime they want.
                    imdb_poster_url_pattern = re.search(
                        regex_url, imdb_movie_cropped_poster_url
                    )
                    imdb_poster_url = (
                        re.sub(regex_url, "", imdb_movie_cropped_poster_url)
                        if imdb_poster_url_pattern
                        else imdb_movie_cropped_poster_url
                    )

                    imdb_year_parentheses_test = re.search(
                        regex_year_parentheses, imdb_title
                    )
                    imdb_year_parentheses = (
                        re.findall(regex_year_parentheses, imdb_title)[0]
                        if imdb_year_parentheses_test
                        else "N/A"
                    )

                    imdb_title_without_parentheses_test = re.search(
                        regex_title, imdb_title
                    )
                    imdb_title_without_parentheses = (
                        re.findall(regex_title, imdb_title)[0]
                        if imdb_title_without_parentheses_test
                        else imdb_title
                    )
                    logging.debug(
                        "The IMDB raw poster URL is: %s" % imdb_movie_cropped_poster_url
                    )
                    logging.debug("The IMDB poster URL is: %s" % imdb_poster_url)
                    logging.debug(
                        "Imdb extracted title without parentheses is: %s"
                        % imdb_title_without_parentheses
                    )

                    try:
                        int(imdb_year_parentheses)
                        is_year_int = True
                    except ValueError:
                        is_year_int = False

                    # If the value between the parentheses is not an integer,
                    # then we extract the first
                    # year value from the text inside the parentheses.
                    if not is_year_int:
                        imdb_year_string_test = re.search(
                            regex_year, imdb_year_parentheses
                        )
                        if imdb_year_string_test:
                            imdb_year_from_string = re.findall(
                                regex_year, imdb_year_parentheses
                            )[0]
                        else:
                            imdb_year_from_string = "N/A"

                    imdb_media_types = {
                        "mov": "movie",
                        "tv": "tvSeries",
                        "tvmini": "tvMiniSeries",
                    }

                    try:
                        imdb_pageconst = item["pagemap"]["metatags"][0][
                            "imdb:pageconst"
                        ]
                    except Exception as e:
                        imdb_pageconst = None

                    imdb_search_criteria = {
                        "media_title": imdb_title_without_parentheses,
                        "media_pageconst": imdb_pageconst,
                        "media_type": imdb_media_types[media_type],
                        "media_year": imdb_year_parentheses
                        if is_year_int
                        else imdb_year_from_string,
                    }

                    imdb_movie_data = imdb_get_data_from_datasets(imdb_search_criteria)

                    # If imdb_get_data_from_datasets() function doesn't return anything,
                    # then we assume that either the requested
                    # movie doesn't exist on the dataset or that the media type was incorrect
                    # (e.g., specifying "tvSeries" instead of "movie" for a movie).
                    try:
                        if imdb_movie_data == None:
                            raise TypeError
                    except TypeError:
                        click.echo(
                            Tcolors.FAIL
                            + "Either the requested title doesn't exist or that"
                            " the media type was incorrectly specified!" + Tcolors.ENDC
                        )
                        logging.warning(
                            "Either the requested title doesn't exist or that"
                            " the media type was incorrectly specified!"
                        )
                        return

                    try:
                        float(imdb_movie_data["averageRating"])
                        is_rating_float = True
                    except (ValueError, TypeError) as e:
                        is_rating_float = False

                    results = {
                        "imdb_title": imdb_movie_data["primaryTitle"],
                        "imdb_original_title": imdb_movie_data["originalTitle"],
                        "imdb_year": imdb_year_parentheses
                        if is_year_int
                        else imdb_year_from_string,
                        "imdb_rating": imdb_movie_data["averageRating"]
                        if is_rating_float
                        else "N/A",
                        "imdb_genres": imdb_movie_data["genres"],
                        "imdb_runtime": imdb_movie_data["runtimeMinutes"]
                        if imdb_movie_data["runtimeMinutes"] != "\\N"
                        else "N/A",
                        "imdb_description": imdb_description,
                        "imdb_thumbnail_url": imdb_movie_thumbnail_url,
                        "imdb_poster_url": imdb_poster_url,
                    }

                    logging.info("Imdb's media data: %s" % results)
                    return results
        else:
            logging.warning("No results!")
            click.echo(Tcolors.WARNING + "No results!" + Tcolors.ENDC)


def rt_get_data(title, title_original, mtype, year):
    """This function uses RT semi-public search API to get information about
    a specific title (movie or series)."""

    movie_title = title
    movie_title_original = title_original
    media_type = mtype
    media_year = year
    rt_media_types = {
        "mov": "movies",
        "tv": "tvSeries",
        "tvmini": "tvSeries",
    }

    try:
        int(media_year)
        is_year_int = True
    except ValueError:
        is_year_int = False

    if not is_year_int:
        return

    def rt_json_fetcher(rt_movie_title):
        """This function uses RT semi-public search API to get a JSON response containing
        information about a specific title (movie or series)."""

        movie_title_rt = rt_movie_title
        query = urllib.parse.quote_plus(str(movie_title_rt))
        url = f"https://www.rottentomatoes.com/api/private/v2.0/search?q={query}"

        try:
            res = urlopen(url)
        except HTTPError as e:
            logging.critical("RT API server couldn't fulfill the request.")
            logging.debug("Error code: %s" % e.code)
            click.echo(
                Tcolors.FAIL
                + "RT API server couldn't fulfill the request."
                + Tcolors.ENDC
            )
        except URLError as e:
            logging.critical("We failed to reach RT API server.")
            click.echo(
                Tcolors.FAIL + "We failed to reach RT API server." + Tcolors.ENDC
            )
            if hasattr(e, "reason"):
                logging.debug("Reason: %s" % e.reason)
        else:
            data = json.loads(res.read().decode())
            media_list_json = data.get(rt_media_types[media_type])
            return media_list_json

    media_list = rt_json_fetcher(movie_title)

    if not media_list:
        rt_search_title = rt_search_media(movie_title, rt_media_types[media_type])
        if not rt_search_title:
            return
        else:
            media_list = rt_json_fetcher(rt_search_title)
            if not media_list:
                logging.warning("RT search returned no results!")
                click.echo(
                    Tcolors.WARNING + "RT search returned no results!" + Tcolors.ENDC
                )
                return

    results = {}
    for item in media_list:
        try:
            if rt_media_types[media_type] == "movies":
                rt_title = item["name"]
                rt_year = item["year"]
            else:
                rt_title = item["title"]
                rt_year = item["startYear"]

            rt_freshness = item["meterClass"]

        except KeyError as err:
            logging.critical("KeyError: {0}".format(err))
            click.echo(Tcolors.FAIL + "KeyError: %s" % err + Tcolors.ENDC)
            continue
        except NameError as err:
            logging.critical("NameError: {0}".format(err))
            click.echo(Tcolors.FAIL + "NameError: %s" % err + Tcolors.ENDC)
        except IndexError as err:
            logging.critical("IndexError: {0}".format(err))
            click.echo(Tcolors.FAIL + "IndexError: %s" % err + Tcolors.ENDC)
        except Exception as e:
            logging.critical("Unexpected error: {0}".format(e))
            click.echo(Tcolors.FAIL + "Unexpected error: %s" % e + Tcolors.ENDC)
            raise
        else:

            # Getting the similarity ratio between IMDb title and RT title
            similar_title_1 = SequenceMatcher(None, rt_title, movie_title).ratio()
            similar_title_2 = SequenceMatcher(
                None, rt_title, movie_title + " " + movie_title_original
            ).ratio()
            logging.debug(
                "Similar ratio 1: %s, similar ratio 2: %s"
                % (similar_title_1, similar_title_2)
            )

            # Sometimes the release year in RT is slightly different than that of IMDb.
            # In this case, it is usually off by 1 or 2 years).
            # Here, we use that difference and combine it with the similar ratio to get
            # the exact title we are looking for.
            # The code here is not bullet-proof and might provide inaccurate results,
            # but this is very rare according to my tests.
            # We can always modify the similarity threshold to increase the accuracy.
            if (
                rt_year == int(media_year)
                or rt_year in range(int(media_year) - 2, int(media_year) + 3)
            ) and max(similar_title_1, similar_title_2) >= 0.75:
                logging.debug("Matched year is: %s" % rt_year)

                rt_rating = item.get("meterScore")
                results = {
                    "rt_title": rt_title,
                    "rt_year": rt_year,
                    "rt_rating": rt_rating if rt_rating else "N/A",
                    "rt_freshness": rt_freshness,
                }

                logging.info("RT's media data: %s" % results)
                return results


def write_json_to_file(json_dict):
    """This functions writes a dictionary (input) to a JSON file."""

    dict_input = json_dict
    with open("results.json", "w") as json_file:
        json.dump(dict_input, json_file, indent=4)


if __name__ == "__main__":
    imdb_cli_init()
    # movie = rt_search_media("westworld", "tvSeries")
    # print("MOVIE", movie)
    # data = rt_get_data("American Horror Story", "American Horror Story", "tv", "2013")
    # print(data)
    # results = rt_parse_json()
    # print(results)
    # rt_construct_json()
    # year = rt_get_movie_year("/m/can_you_ever_forgive_me")
    # print(year)
    # imdb_get_data("arrows")



from pathlib import Path
import pandas as pd
import logging
import click
import datetime
from imgdb.utils import Tcolors, unzip, pickler
from imgdb.config import Config
from imgdb.imdb_poster_fetcher import imdb_download_poster
import re


def datasets_updater(freq):
    """This functions downloads Imdb's datasets and updates
    them according to a user-defined interval."""

    title_basics_url = "https://datasets.imdbws.com/title.basics.tsv.gz"
    title_ratings_url = "https://datasets.imdbws.com/title.ratings.tsv.gz"
    datasets_list = [title_basics_url, title_ratings_url]
    regex_tsv_title = r"(?<=com/)(.*?)(?=.gz)"
    tsv_time_saved = {}
    date_now = datetime.datetime.now()
    date_test = datetime.datetime(2021, 11, 24, 23, 59, 59)
    tsv_files = []
    tsv_save_pickle_path = Path(Config.IMDB_DATASETS_DIR / "tsv_save.pickle")
    regex_freq = r"(^[0-9]{1,}h|d)$"
    freq_test = re.search(regex_freq, freq)
    is_pickle = tsv_save_pickle_path.is_file()

    if is_pickle:
        saved_pickle = pickler(tsv_save_pickle_path)
        time_difference = (date_now - saved_pickle["time"]).total_seconds()
        logging.debug("Time difference is: %s" % time_difference)
        logging.debug("Content of pickle file: %s" % saved_pickle)

    update_freq = {"daily": 86400, "weekly": 604800, "bi-weekly": 1209600}

    count_freq = 0
    for key, value in update_freq.items():
        count_freq += 1
        if key == freq:
            threshold = update_freq[key]
            break
        elif freq_test:
            if freq[-1:] == "h":
                threshold = int(freq[:-1]) * 3600
            elif freq[-1:] == "d":
                threshold = int(freq[:-1]) * 86400
            break
        elif count_freq == len(update_freq):
            logging.warning(
                "The update frequency format is wrong." " Using 'weekly' as a fallback!"
            )
            click.echo(
                Tcolors().WARNING + "The update frequency format is wrong."
                " Using 'weekly' as a fallback!" + Tcolors.ENDC
            )
            threshold = update_freq["weekly"]
            logging.debug("Threshold inside loop is %s" % threshold)
        else:
            continue

    logging.debug("The threshold is %s" % threshold)

    count_url = 0
    for url in datasets_list:
        count_url += 1
        file_name_url_test = re.search(regex_tsv_title, url)
        if file_name_url_test:
            tsv_file_name = re.findall(regex_tsv_title, url)[0]
            tsv_gz_file_name = url.split("/")[-1]
            tsv_gz_file_path = Path(Config.IMDB_DATASETS_DIR / tsv_gz_file_name)
            tsv_file_path = Path(Config.IMDB_DATASETS_DIR / tsv_file_name)
            tsv_files.append(tsv_file_path)
            logging.debug("The filename of the tsv file is: %s" % tsv_file_name)
            is_file = tsv_file_path.is_file()

            if (not is_file or not is_pickle) or (
                is_pickle and time_difference > threshold
            ):
                download_msg = "Downloading" if not is_pickle else "Updating"
                logging.info("Downloading dataset file %s... yay :)" % tsv_gz_file_name)
                if count_url == 1:
                    click.echo(
                        Tcolors.OK_GREEN
                        + "%s IMDb datasets!" % download_msg
                        + Tcolors.ENDC
                    )
                # Downloading IMDb datasets
                imdb_download_poster(
                    url,
                    name=tsv_file_name,
                    filepath=tsv_gz_file_path,
                    download_or_update=download_msg,
                )

                click.echo(
                    Tcolors.OK_GREEN + "Unzipping: %s" % tsv_gz_file_name + Tcolors.ENDC
                )
                logging.info("Unzipping: %s" % tsv_gz_file_name)
                # Unzipping the datsets
                unzip(tsv_gz_file_path, tsv_file_path)

                if count_url == len(datasets_list):
                    tsv_time_saved["time"] = date_now
                    logging.debug("Time inside pickle file is: %s" % date_now)
                    pickler(tsv_save_pickle_path, tsv_time_saved)

                    logging.debug("The tsv list: %s" % tsv_files)
                    merge_tsv_files(
                        tsv_files[0], tsv_files[1], Config.IMDB_DATASETS_DIR
                    )
                    logging.debug("The tsv files were merged successfully!")


def merge_tsv_files(tsv_title_basics, tsv_title_ratings, base_path):
    """This function merges two tsv files."""

    title_basics_ratings_file_path = Path(base_path / "title_basics_ratings.tsv")
    try:
        title_basics_pd = pd.read_csv(
            tsv_title_basics,
            sep="\t",
            header=0,
            dtype={
                "tconst": str,
                "titeType": str,
                "primaryTitle": str,
                "originalTitle": str,
                "isAdult": str,
                "startYear": str,
                "endYear": str,
                "runtimeMinutes": str,
                "genres": str,
            },
        )
        title_ratings_pd = pd.read_csv(
            tsv_title_ratings,
            sep="\t",
            header=0,
            dtype={"tconst": str, "averageRating": float, "numVotes": int},
        )
    except Exception as e:
        logging.critical("We couldn't read the tsv files!")
        logging.debug("Error: %s" % e)
        click.echo(Tcolors.FAIL + "We couldn't read the tsv files!" + Tcolors.ENDC)
        return 0

    # Merging the two tsv files
    click.echo(Tcolors.OK_GREEN + "Merging IMDb datasets!" + Tcolors.ENDC)
    logging.info("Merging IMDb datasets!")
    try:
        title_basics_ratings_pd = title_basics_pd.merge(title_ratings_pd, on="tconst")
    except Exception as e:
        logging.critical("We failed to merge the tsv files!")
        logging.debug("Error: %s" % e)
        click.echo(Tcolors.FAIL + "We failed to merge the tsv files!" + Tcolors.ENDC)
        return 0
    else:
        click.echo(
            Tcolors.OK_GREEN + "Successfully merged IMDb datasets!" + Tcolors.ENDC
        )
        logging.info("Successfully merged IMDb datasets!")

    # Saving the merged file to disk
    click.echo(
        Tcolors.OK_GREEN + "Saving the merged dataset file to disk!" + Tcolors.ENDC
    )
    logging.info("Saving the merged dataset file to disk!")
    try:
        title_basics_ratings_pd.to_csv(
            title_basics_ratings_file_path, sep="\t", index=False
        )
    except Exception as e:
        logging.critical("We failed to write the merged tsv file!")
        logging.debug("Error: %s" % e)
        click.echo(
            Tcolors.FAIL + "We failed to write the merged tsv file!" + Tcolors.ENDC
        )
        return 0
    else:
        click.echo(
            Tcolors.OK_GREEN
            + "Merged dataset file saved as:\n <%s>" % title_basics_ratings_file_path
            + Tcolors.ENDC
        )
        logging.info(
            "Merged dataset file saved as: <%s>" % title_basics_ratings_file_path
        )


def imdb_get_data_from_datasets(criteria={}):
    """This function takes in a dictionary containing movie data
    as input to output the corresponding Imdb's data.

    The input dictionary has the following structure:
        criteria = {
                        "media_title": The title of the movie,
                        "media_pageconst": The ID of the title,
                        "media_type": (e.g., movie, tvSeries, etc.),
                        "media_year": The release year of the movie,
                    }
    """

    tsv_title_basics_ratings = Path(
        Config.IMDB_DATASETS_DIR / "title_basics_ratings.tsv"
    )

    argument_list = criteria

    chunk_size = 100000
    dtypes = {
        "tconst": str,
        "titeType": str,
        "primaryTitle": str,
        "originalTitle": str,
        "isAdult": str,
        "startYear": str,
        "endYear": str,
        "runtimeMinutes": str,
        "genres": str,
        "averageRating": float,
        "numVotes": int,
    }

    for chunk in pd.read_csv(
        tsv_title_basics_ratings, sep="\t", chunksize=chunk_size, dtype=dtypes, header=0
    ):

        if argument_list["media_pageconst"]:
            res = chunk.loc[
                (chunk["tconst"] == argument_list["media_pageconst"])
                & (chunk["titleType"] == argument_list["media_type"])
            ]
        else:
            res = chunk.loc[
                (chunk["primaryTitle"] == argument_list["media_title"])
                & (chunk["titleType"] == argument_list["media_type"])
                & (chunk["startYear"] == argument_list["media_year"])
            ]

        if not res.empty:
            imdb_movie_raw_data = res.values[0]
            imdb_movie_data = {
                "tconst": imdb_movie_raw_data[0],
                "titleType": imdb_movie_raw_data[1],
                "primaryTitle": imdb_movie_raw_data[2],
                "originalTitle": imdb_movie_raw_data[3],
                "isAdult": imdb_movie_raw_data[4],
                "startYear": imdb_movie_raw_data[5],
                "endYear": imdb_movie_raw_data[6],
                "runtimeMinutes": imdb_movie_raw_data[7],
                "genres": imdb_movie_raw_data[8].split(","),
                "averageRating": imdb_movie_raw_data[9],
                "numVotes": imdb_movie_raw_data[10],
            }

            logging.debug("Imdb movie data: %s" % imdb_movie_data)
            logging.info("The results were successfully retrieved from the datasets.")
            return imdb_movie_data


if __name__ == "__main__":
    imdb_search_criteria = {
        "movie_title": "Rams",
        "media_type": "movie",
        "movie_year": "2015",
    }
    print(imdb_get_data_from_datasets(imdb_search_criteria))



from wand.image import Image
from wand.color import Color
from wand.drawing import Drawing
from wand.display import display
from wand.font import Font
import click
from imgdb.utils import Tcolors
from imgdb.config import Config
import textwrap
import logging
from importlib.resources import files
import imgdb.resources.images as img_dir


def generate_media_image(
    movie_title,
    imdb_rating,
    rt_rating,
    poster_filename,
    poster_filepath,
    font=None,
):

    if poster_filename:
        saved_image_filename = poster_filename + "_by_imgdb"
    else:
        click.echo(
            Tcolors.WARNING + "Downloaded image filename"
            " cannot be null!" + Tcolors.ENDC
        )
        logging.critical("Downloaded image filename cannot be null!")
        return

    try:
        float(imdb_rating)
    except (ValueError, TypeError):
        imdb_is_rating_float = False
    else:
        imdb_is_rating_float = True
        imdb_rating_to_perc = (
            imdb_rating if isinstance(imdb_rating, float) else float(imdb_rating)
        ) * 10

    try:
        float(rt_rating)
    except (ValueError, TypeError):
        rt_is_rating_int = False
    else:
        rt_is_rating_int = True
        rt_rating_to_int = rt_rating if isinstance(rt_rating, int) else int(rt_rating)

    if not imdb_is_rating_float and not rt_is_rating_int:
        click.echo(
            Tcolors.WARNING + "Cannot produce picture"
            " without ratings!" + Tcolors.ENDC
        )
        logging.critical("Cannot produce picture without ratings!")
        return

    else:
        if imdb_is_rating_float:
            imdb_rounded_rating = round_ratings(imdb_rating_to_perc)
            logging.debug("IMDb's rounded rating is: %s" % imdb_rounded_rating)
        if rt_is_rating_int:
            rt_rounded_rating = round_ratings(rt_rating_to_int)
            logging.debug("RT's rounded rating is: %s" % rt_rounded_rating)

    # Image processing
    with Image(width=800, height=800, background="#3D3737") as canvas:
        with Drawing() as context:
            # Poster image
            with Image(filename=poster_filepath) as poster:
                poster_width = 380
                poster.transform(resize=f"{poster_width}x")
                while poster.height > 570:
                    poster_width -= 5
                    poster.transform(resize=f"{poster_width}x")
                logging.debug("Poster's final height %s" % poster.height)
                canvas.composite(poster, left=40, top=180)

            # Ring charts
            ## Charts' parameters
            ring_size = 140
            x_ring_chart = 460
            y_imdb_rating = 200
            y_rt_rating = 440
            x_rating_text_offset = 130
            y_rating_text_offset = 170
            x_shadow_text_offset = 7
            y_shadow_text_offset = 7

            ## IMDb ring chart rating
            if imdb_is_rating_float:
                imdb_rating_img = files(img_dir).joinpath(
                    f"imdb_{imdb_rounded_rating}.png"
                )
                with Image(filename=imdb_rating_img) as img_imdb:
                    img_imdb.resize(ring_size, ring_size)
                    canvas.composite(img_imdb, left=x_ring_chart, top=y_imdb_rating)

            ## RT ring chart rating
            if rt_is_rating_int:
                rt_rating_img = files(img_dir).joinpath(f"rt_{rt_rounded_rating}.png")
                with Image(filename=rt_rating_img) as img_rt:
                    img_rt.resize(ring_size, ring_size)
                    canvas.composite(img_rt, left=x_ring_chart, top=y_rt_rating)

            # Text
            x_title_text = 25
            x_shadow_title_offset = 5
            y_shadow_title_offset = 5
            fontsize_title = 50
            fontsize_ratings = 70
            # context.font = "Mosk-Ultra-Bold-900"
            if font:
                context.font = font
            else:
                context.font = "Maler-Regular"

            header_text = textwrap.fill(movie_title, width=15)
            font_metrics = context.get_font_metrics(
                canvas, header_text.upper(), multiline=True
            )
            logging.debug("Initial font metrics: %s" % str(font_metrics))

            while font_metrics.text_width > 600 or font_metrics.text_height > 150:
                context.font_size -= 1
                font_metrics = context.get_font_metrics(
                    canvas, header_text.upper(), multiline=True
                )
            else:
                while font_metrics.text_width < 600 and font_metrics.text_height < 150:
                    context.font_size += 1
                    font_metrics = context.get_font_metrics(
                        canvas, header_text.upper(), multiline=True
                    )
            logging.debug("Corrected font metrics: %s" % str(font_metrics))

            y_title = int(
                ((180 - font_metrics.text_height) / 2) + font_metrics.character_height
            )
            logging.debug("Title's y-axis value: %s" % y_title)
            context.push()

            ## Text shadow
            context.fill_color = Color("rgba(29, 29, 29, 0.87)")
            context.text(
                x=x_title_text + x_shadow_title_offset,
                y=y_title + y_shadow_title_offset,
                body=header_text.upper(),
            )
            context.font_size = fontsize_ratings
            if imdb_is_rating_float:
                context.text(
                    x=x_ring_chart + x_rating_text_offset + x_shadow_text_offset,
                    y=y_imdb_rating + y_rating_text_offset + y_shadow_text_offset,
                    body=str(imdb_rating),
                )
            if rt_is_rating_int:
                context.text(
                    x=x_ring_chart + x_rating_text_offset + x_shadow_text_offset,
                    y=y_rt_rating + y_rating_text_offset + y_shadow_text_offset,
                    body=str(rt_rating) + "%",
                )

            ## Main text
            context.pop()
            context.fill_color = Color("white")
            context.text(x=x_title_text, y=y_title, body=header_text.upper())
            context.font_size = fontsize_ratings
            if imdb_is_rating_float:
                context.text(
                    x=x_ring_chart + x_rating_text_offset,
                    y=y_imdb_rating + y_rating_text_offset,
                    body=str(imdb_rating),
                )
            if rt_is_rating_int:
                context.text(
                    x=x_ring_chart + x_rating_text_offset,
                    y=y_rt_rating + y_rating_text_offset,
                    body=str(rt_rating) + "%",
                )

            # Text stroke (Obsolete)
            # context.fill_color = Color("transparent")
            # context.stroke_color = Color("rgba(29, 29, 29, 0.87)")
            # context.stroke_width = 1
            # context.text(x=21, y=75, body=movie_title.upper())
            # context.font_size = fontsize_ratings
            # if imdb_is_rating_float:
            #     context.text(x=485, y=290, body=str(imdb_rating))
            # if rt_is_rating_int:
            #     context.text(x=485, y=510, body=str(rt_rating) + "%")

            context(canvas)
            canvas.format = "png"
            if Config.DEV_MODE:
                display(canvas)
        canvas.save(filename=f"{saved_image_filename}.png")
        logging.info("Edited image saved as: %s" % f"{saved_image_filename}.png")
        click.echo(
            Tcolors.OK_GREEN
            + "Edited image saved as: %s" % f"{saved_image_filename}.png"
            + Tcolors.ENDC
        )


def round_ratings(rating):
    """This functions takes in a movie rating as an argument
    and returns its corresponding rounded value
    that can be used to choose the correct ring chart picture."""

    rounded_rating_value = None
    rating_steps = [0, 5, 10, 20, 25, 30, 40, 50, 60, 70, 75, 80, 90, 95, 100]
    for step in rating_steps:
        if rating in rating_steps:
            rounded_rating_value = rating_steps[rating_steps.index(rating)]
            break
        # Using zero or 100 only when the rating exactly matches those values.
        # In other words, we don't round values like 2 or 98 for example
        # to zero or 100 respectively!
        elif rating in range(1, 3):
            rounded_rating_value = 5
            break
        elif rating in range(98, 100):
            rounded_rating_value = 95
            break
        else:
            rating_diff = rating - step
            rating_diff_abs = abs(rating_diff)

            if (
                rating in range(3, 10)
                or rating in range(21, 30)
                or rating in range(71, 80)
                or rating in range(91, 100)
            ):
                difference_threshold = 2.5
            else:
                difference_threshold = 5
                rounding_increment = 10

            if rating_diff_abs < difference_threshold:
                rounded_rating_value = step
                break
            elif (
                rating_diff_abs == difference_threshold and difference_threshold != 2.5
            ):
                rounded_rating_value = step + rounding_increment
                break

    logging.debug("Rounded rating value is: %s" % rounded_rating_value)
    return rounded_rating_value


if __name__ == "__main__":
    generate_media_image(
        movie_title="The forever purge",
        imdb_rating="5.4",
        rt_rating=0,
        poster_filename="The forever purge",
        poster_filepath="Meander.jpg",
        font=None,
    )
    rt_base_url = "fag§"
    api_urls = [
        f"{rt_base_url}/browse?minTomato=70&maxTomato=100"
        "&maxPopcorn=100&services=amazon%3Bhbo_go%3B"
        "itunes%3Bnetflix_iw%3Bvudu%3Bamazon_prime%3Bfandango_now"
        "&certified=true&sortBy=release&type=cf-dvd-streaming-all",
        f"{rt_base_url}/browse?minTomato=70&maxTomato=100"
        "&maxPopcorn=100&services=amazon%3Bhbo_go%3B"
        "itunes%3Bnetflix_iw%3Bvudu%3Bamazon_prime%3Bfandango_now"
        "&certified=true&sortBy=release&type=cf-dvd-streaming-all&page=2",
    ]
    print(api_urls[0])



import logging
import pickle
import gzip
import shutil
import click
import sys


def logger(loglevel, filepath):
    """This function serves as a logger."""

    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        click.echo(
            Tcolors.WARNING
            + "Invalid log level: %s. Using 'warning' as a fallback!" % loglevel
            + Tcolors.ENDC
        )
        numeric_level = 30
    logging.basicConfig(
        level=numeric_level,
        format="%(asctime)s - %(name)s - %(levelname)-8s"
        " [%(filename)s:%(lineno)d] %(message)s",
        datefmt="%m/%d/%Y %I:%M:%S %p",
        filename=filepath,
    )


class Tcolors:
    """A set of ANSI colors."""

    HEADER = "\033[95m"
    OK_BLUE = "\033[94m"
    OK_CYAN = "\033[96m"
    OK_GREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def unzip(filepath_input, filepath_output):
    """This function uses gzip to unpack a gzipped file."""

    try:
        with gzip.open(filepath_input, "rb") as gz_in:
            with open(filepath_output, "wb") as tsv_out:
                shutil.copyfileobj(gz_in, tsv_out)
    except Exception as e:
        logging.critical("Something went wrong while trying to open the gzipped files!")
        logging.debug("Error: %s" % e)
        click.echo(
            Tcolors.FAIL
            + "Something went wrong while trying to open the gzipped files!"
            + Tcolors.ENDC
        )
        sys.exit()

    try:
        filepath_input.unlink()
    except FileNotFoundError:
        logging.warning("Couldn't delete %s: The file doesn't exist!" % filepath_input)
    except Exception as e:
        logging.critical(
            "Something went wrong while trying to delete the gzipped files!"
        )
        logging.debug("Error: %s" % e)
        click.echo(
            Tcolors.FAIL
            + "Something went wrong while trying to delete the gzipped files!"
            + Tcolors.ENDC
        )
        sys.exit()


def pickler(save_pickle_path, save_pickle_input=None):
    """This function uses the pickle library to serialize
    or de-serialize an input dictionary."""

    if save_pickle_input:
        try:
            with open(save_pickle_path, "wb") as handle:
                pickle.dump(save_pickle_input, handle, protocol=pickle.HIGHEST_PROTOCOL)
        except Exception as e:
            logging.critical(
                "Something went wrong while trying to save the program's state!"
            )
            logging.debug("Error: %s" % e)
            click.echo(
                Tcolors.FAIL
                + "Something went wrong while trying to save the program's state!"
                + Tcolors.ENDC
            )
    else:
        try:
            with open(save_pickle_path, "rb") as read_pickle:
                unpickled_output = pickle.load(read_pickle)
                return unpickled_output
        except Exception as e:
            logging.critical(
                "Something went wrong while trying to read the program's state!"
            )
            logging.debug("Error: %s" % e)
            click.echo(
                Tcolors.FAIL
                + "Something went wrong while trying to read the program's state!"
                + Tcolors.ENDC
            )
            sys.exit()


def replace_every_nth(nth, pattern, drop, string):
    """This function replaces every nth character (pattern) in a string
    with a pre-defined text or character (drop)."""

    string_to_list = list(string)
    cursor = 0
    for idx in range(len(string_to_list)):
        if string_to_list[idx] == pattern:
            if len(string_to_list[cursor + 1 : idx]) >= nth:
                string_to_list.insert(idx + 1, drop + "             ")
                cursor = idx

    list_to_string = "".join(string_to_list)
    return list_to_string


if __name__ == "__main__":
    string = "True Detective: Created by Nic Pizzolatto. With Matthew McConaughey, Colin Farrell, Mahershala Ali, Woody Harrelson. Seasonal anthology series in which police investigations unearth the personal and professional secrets of those involved, both within and outside the law."
    s = replace_every_nth(50, " ", "\n", string)
    print(s)
